from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from types import SimpleNamespace

import numpy as np

def step0_args(method: str, *args, **kwargs) -> dict:
    # Loyvain arguments initialization

    n_args = len(args)
    if n_args < 2:
        raise ValueError("Wrong number of input arguments.")
    match method:
        case "loyvain":
            n_args_num = 2
            W, k = args[:n_args_num]
            X = W
            Y = np.array(None)
        case "coloyvain":
            n_args_num = 2 + (n_args > 2 and (not isinstance(args[2], str)))
            if n_args_num == 2:
                W, k = args[:n_args_num]
                X, Y = np.array(None), np.array(None)
            elif n_args_num == 3:
                X, Y, k = args[:n_args_num]
                W = np.array(None)
    args = args[n_args_num:]
    if n_args >= n_args_num + 1:
        kwargs["objective"] = args[0]
        args = args[1:]
        if n_args >= n_args_num + 2:
            kwargs["similarity"] = args[0]
            args = args[1:]

    Args = parse_args(method=method, W=W, X=X, Y=Y, k=k, **kwargs)

    return SimpleNamespace(**Args)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def parse_args(
    method: Literal["loyvain", "coloyvain"],
    W: ArrayLike = None,
    X: ArrayLike = None,
    Y: ArrayLike = None,
    k: int = 0,
    objective: Literal["kmodularity", "kmeans", "spectral"] = "kmodularity",
    similarity: Literal["network", "corr", "cosim", "cov", "dot"] = "network",
    start: ArrayLike | Literal["greedy", "balanced", "random"] = "greedy",
    numbatches: int = 10,
    maxiter: int = 1000,
    replicates: int = 10,
    tolerance: float = 1e-10,
    display: Literal["none", "replicate", "iteration"] = "none",
) -> dict:
    
    W = np.asarray(W)
    X = np.asarray(X)
    Y = np.asarray(Y)

    if method == "coloyvain":
        if isinstance(start, np.ndarray):
            raise ValueError("Start cannot be a numeric vector for co-Loyvain.")
        if k <= 0:
            raise ValueError("k must be positive for co-Loyvain.")
        if similarity == "network":
            if not (np.array_equal(X, None) and np.array_equal(Y, None)):
                raise ValueError('X and Y inputs are incompatible with "network" similarity.')
        else:
            if not np.array_equal(W, None):
                raise ValueError('W input is only compatible with "network" similarity.')
            if X.shape[0] != Y.shape[0]:
                raise ValueError("X and Y must have the same number of data points.")

    return {key: value for key, value in locals().items()}



import numpy as np

def step3_init(X, normX, Dist, n, Args):
    # Loyvain partition initialization

    # Unpack arguments
    k = Args.k

    if Args.start in ["greedy", "balanced"]:
        Idx = np.concatenate(([np.random.randint(n)], np.full(k - 1, -1)))
        minDist = np.full(n, np.inf)
        for j in range(1, k):
            if (Args.similarity == "network") or (Args.method == "coloyvain"):
                # use precomputed distance
                Dj = Dist[Idx[j-1]]
            else:
                # compute distance on the fly
                Dj = 1 - (X[Idx[j-1]] @ X.T) / (normX[Idx[j-1]] * normX.T)
            minDist = np.minimum(minDist, Dj)  # min distance to centroid
            if Args.start == "greedy":
                sampleProbability = (minDist == np.max(minDist))
            elif Args.start == "balanced":
                sampleProbability = (minDist / np.sum(minDist))
            P = np.concatenate(([0], np.cumsum(sampleProbability)))
            P[-1] = 1
            Idx[j] = np.searchsorted(P, np.random.rand()) - 1 # sample next centroid

        if (Args.similarity == "network") or (Args.method == "coloyvain"):
            # use precomputed distance
            M0 = np.argmin(Dist[Idx], axis=0)
        else:
            # compute distance on the fly
            M0 = np.argmin(1 - (X[Idx] @ X.T) / (normX[Idx] * normX.T), axis=0)
        k0 = np.max(M0) + 1
        if k0 < k:
            M0[np.random.choice(n, k - k0, replace=False)] = np.arange(k0, k)
    elif Args.start == "random":
        M0 = np.random.randint(k, size=n)  # initial module partition
        M0[np.random.choice(n, k, replace=False)] = np.arange(k)  # ensure there are k modules

    return M0



```
 KNEIGHBOR Common-neighbor or symmetric kappa-nearest-neighbor matrix

   B = kneighbor(W)
   B = kneighbor(W, type)
   B = kneighbor(W, type, kappa)
   B = kneighbor(X, type, kappa, similarity)
   B = kneighbor(X, type, kappa, similarity, method)
   B = kneighbor(X, type, kappa, similarity, method, Name=Value)

   Inputs:

       W: Network matrix of size n x n.
       OR
       X: Data matrix of size n x p, where
           n is the number of data points and
           p is the number of features.

       type: Type of neighbor matrix.
           "common": Common-neighbor matrix (default).
           "nearest": Symmetric kappa-nearest neighbor matrix.

       kappa: Number of nearest neighbors.
           1 <= kappa < n (default is 10).
           OR
           0 < kappa < 1 to use as a fraction of n.

       similarity: Type of similarity.
           "network": Network connectivity (default).
           "corr": Pearson correlation coefficient.
           "cosim": Cosine similarity.

       method: Method of neighbor search.
           "direct": Direct computation of similarity matrix (default).
           "indirect": knnsearch (in MATLAB)
                       pynndescent (in Python).

       Name=Value Arguments:
           Optional arguments passed to knnsearch or pynndescent.

   Outputs:
       B: Co-neighbor or symmetric nearest-neighbor matrix (size n x n).

   Methodological notes:
       Symmetric kappa-nearest-neighbor matrices are binary matrices that
       connect pairs of nodes if one of the nodes is a top-kappa nearest
       neighbor of the other node (in a structural, correlation, or
       another network).

       kappa-common-neighbor matrices are symmetric integer matrices that
       connect pairs of nodes by the number of their shared top-kappa
       nearest neighbors.
 
       Direct computation of the similarity matrix is performed in
       blocks. It is generally faster than indirect computation.

   Dependencies:
       MATLAB: 
           Statistics and Machine Learning Toolbox (if method="indirect")
       Python: 
           PyNNDescent (if method="indirect")

   See also:
       KNEICOMP, MUMAP.

```



```
 COLOYVAIN K-modularity, k-means, or spectral co-clustering

   Mx, My, R = coloyvain(W, k)
   Mx, My, R = coloyvain(W, k, objective)
   Mx, My, R = coloyvain(X, Y, k, objective, similarity)
   Mx, My, R = coloyvain(_, objective, similarity, Name=Value)

   Inputs:

       W: Bipartite network matrix of size p x q.

       X: Data matrix of size n x p, where
          n is the number of data points and
          p is the number of features.

       Y: Data matrix of size n x q, where
          n is the number of data points and
          q is the number of features.

       k: Number of modules (positive integer).

       objective: Clustering objective.
           See LOYVAIN for all options.

       similarity: Type of similarity.
           See LOYVAIN for all options.

       Name=Value Arguments.
           See LOYVAIN for all Name=Value arguments.

   Outputs:
       Mx: Vector of module assignments for X (length p).
       My: Vector of module assignments for Y (length q).
       R: Value of maximized objective.

   Methodological notes:
       Coloyvain simultaneously clusters X and Y via Loyvain
       co-clustering of the cross-similarity matrix.

   See also:
       LOYVAIN, CANONCOV.

```



"""
all abct functions
"""

from abct.canoncov import canoncov
from abct.coloyvain import coloyvain
from abct.degree import degree
from abct.dispersion import dispersion
from abct.kneicomp import kneicomp
from abct.kneighbor import kneighbor
from abct.leiden import leiden
from abct.loyvain import loyvain
from abct.mumap import mumap
from abct.residualn import residualn
from abct.shrinkage import shrinkage

__all__ = [
    "canoncov",
    "coloyvain",
    "degree",
    "dispersion",
    "kneicomp",
    "kneighbor",
    "leiden",
    "loyvain",
    "mumap",
    "residualn",
    "shrinkage",
]



from typing import Literal
from numpy.typing import ArrayLike
from scipy.sparse import sparray as SparseArray
from pydantic import validate_call, ConfigDict
from types import SimpleNamespace

import torch
import numpy as np

def step0_args(X: ArrayLike, **kwargs) -> SimpleNamespace:
    # m-umap arguments initialization

    Args = parse_args(X=X, **kwargs)

    return SimpleNamespace(**Args)

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def parse_args(
    X: ArrayLike | SparseArray,
    d: int = 3,
    kappa: float = 30,
    alpha: float = 1.0,
    beta: float = 1.0,
    gamma: float = 1.0,
    similarity: Literal["network", "corr", "cosim"] = "network",
    method: Literal["direct", "indirect"] = "direct",
    replicates: int = 10,
    finaltune: bool = True,
    partition: ArrayLike = None,
    start: ArrayLike | Literal["greedy", "spectral", "spectral_nn"] = "greedy",
    solver: Literal["adam", "trustregions"] = "adam",
    maxiter: int = 10000,
    learnrate: float = 1e-3,
    tolerance: float = 1e-6,
    gpu: bool = False,
    cache: bool = False,
    verbose: bool = True,
) -> dict:

    X = np.asarray(X)
    if similarity == "network":
        if (X.shape[0] != X.shape[1]) or not np.allclose(X, X.T):
            raise ValueError('Network matrix must be symmetric or similarity must not be "network".')

    if gpu and not torch.cuda.is_available():
        raise ValueError("GPU must be available or gpu must be False.")

    return {key: value for key, value in locals().items()}



from numpy.typing import ArrayLike

import numpy as np
from abct import muma

def rotation(U3: ArrayLike) -> tuple[np.ndarray, np.ndarray]:

    # ensure spherical geometry
    n = len(U3)
    U3 = U3 / np.linalg.norm(U3, axis=1, keepdims=True)

    ## Rotate data to empty poles

    p = 144                     # Fibonacci number
    V = muma.fsphere(p)         # Unit Fibonacci sphere

    # Poles have minimal (maximal correlations to all other points)
    D = np.max(np.abs(U3 @ V.T), axis=0)
    idx = np.argmin(D)

    # Rotate data so that v becomes [0, 0, 1]
    v = V[idx]
    Rp = rota(v, 2)
    U3 = U3 @ Rp

    ## Rotate data to sparse anti-prime meridian

    # Unit circle
    Theta = np.pi * np.arange(-p, p, 2)/p
    V = np.column_stack((np.cos(Theta), np.sin(Theta)))

    # Sparse meridians have few strong correlations to other points
    alpha = 12*np.pi/p;         # degree band (6/p in either direction)
    D = np.zeros(p)
    for i in range(p):
        Vi = np.hstack((V[np.full(n, i)], U3[:, 2:3]))
        Vi = Vi / np.linalg.norm(Vi, axis=1, keepdims=True)
        # Number of points within the angle boundary
        D[i] = np.mean(np.sum(U3 * Vi, axis=1) > np.cos(alpha))

    # Find middle meridian among several identical
    Comps = np.hstack((D, D)) == np.min(D)         # Enforce circular boundary conditions
    Comps = np.hstack((0, Comps, 0))
    Comps = np.diff(Comps)
    Comp_sta = np.where(Comps ==  1)[0]
    Comp_fin = np.where(Comps == -1)[0]
    u = np.argmax(Comp_fin - Comp_sta)
    idx = (Comp_sta[u] + Comp_fin[u]) // 2
    v = np.hstack((V[idx % p], 0))

    # Rotate U3 so that v is placed on [-1, 0, 0]
    Rm = rota(-v, 0)
    U3 = U3 @ Rm

    # Combine rotations
    R = Rp @ Rm

    return U3, R

def rota(v: ArrayLike, a: int) -> np.ndarray:
    # Rotate data such that v aligns with a basis vector
    epss = np.finfo(np.float32).eps

    e = np.zeros(3)
    e[a] = 1
    b = np.where(e == 0)[0][0]
    if np.linalg.norm(np.abs(e) - np.abs(v)) < epss:    # Target vector parallel
        R = np.eye(3)
        if np.linalg.norm(e - (- v)) < epss:            # Target vector antiparallel
            R[a, a] = -1
            R[b, b] = -1
    else:
        # Cross product
        x = np.cross(e, v)
        cos_theta = np.dot(e, v)
        sin_theta = np.linalg.norm(x)

        # Skew-symmetric matrix
        Q = np.array([[0, -x[2], x[1]],
                      [x[2], 0, -x[0]],
                      [-x[1], x[0], 0]])

        # Rodrigues formula
        R = np.eye(3) + Q + Q @ Q * ((1 - cos_theta)/(sin_theta**2))

    # Check that rotation matrix is valid and rotation is correct
    assert((np.linalg.norm(np.eye(3) - R.T @ R) < epss) and (np.abs(1 - np.linalg.det(R)) < epss))
    assert(np.linalg.norm(e - v @ R) < epss)

    return R



from typing import Tuple
from importlib.resources import files

import numpy as np
from . import muma

def mumap(*args, **kwargs) -> Tuple[np.ndarray, float]:

    # Parse, process, and test arguments
    Args = muma.step0_args(*args, **kwargs)
    Args = muma.step1_proc(Args)
    muma.step2_test(Args)

    # Initialize and run algorithm
    U = muma.step3_init(Args)
    U, CostHistory = muma.step4_run(U, Args)

    return U, Args.partition, CostHistory

mumap.__doc__ = files("abct").joinpath("docstrings", "doc_mumap.py").read_text().replace("```", "")



```
 KNEICOMP Components of neighbor matrices

   V = kneicomp(W, k)
   V = kneicomp(X, k)
   V = kneicomp(_, k, weight)
   V = kneicomp(_, k, weight, Name=Value)

   Inputs:
       W: Network matrix of size n x n.
       OR
       X: Data matrix of size n x p, where
           n is the number of data points and
           p is the number of features.

       k: Number of components.

       weight: Type of components
           "weighted": Weighted components (default).
           "binary": Binary components.

       Name=Value Arguments:
           KNEIGHBOR: type, kappa, similarity, method
               (see KNEIGHBOR for details).
           LOYVAIN: All Name=Value arguments
               (binary components only, see LOYVAIN for details).

   Outputs:
       V: Component matrix (size n x k).

   Methodological notes:
       By default, weighted components are eigenvectors of
       common-neighbors matrices. In imaging neuroscience, these
       components are approximately equivalent to co-activity gradients
       (diffusion-map embeddings).
 
       Correspondingly, binary components are modules of common-neighbors 
       matrices, estimated using the Loyvain algorithm. They are
       equivalent to eigenvectors of common-neighbors matrices with binary
       constraints. The order of binary components will be arbitrary. 

   See also:
       KNEIGHBOR, LOYVAIN.

```



import numpy as np

def fsphere(p: int) -> np.ndarray:
    # Unit Fibonacci sphere
    I = np.arange(p) + 0.5
    golden_ratio = (1 + np.sqrt(5))/2
    Phi = np.acos(1 - 2*I/p)
    Theta = 2*np.pi*I / golden_ratio
    V = np.column_stack((
            np.cos(Theta)*np.sin(Phi),
            np.sin(Theta)*np.sin(Phi),
            np.cos(Phi)))

    return V



```
 LOUVAINS Efficient Louvain modularity maximization of sparse networks (MATLAB)
 LEIDEN igraph Leiden modularity maximization (Python)

   M, Q = louvains(W, Name=Value)        % MATLAB
   M, Q = leiden(W, Name=Value)          # Python

   Inputs:
       W:  Network matrix of size n x n.

       Name=Value Arguments:

           gamma=Resolution parameter.
               Positive scalar (default is 1).

           start=Initial module assignments.
               Vector of length n (default is 1:n).

           replicates=Number of replicates.
               Positive integer (default is 10).

           finaltune=Final tuning of optimized assignment.
               Logical (default is false).

           tolerance=Convergence tolerance.
               Positive scalar (default is 1e-10).

           display=Display progress.
               "none": no display (default).
               "replicate": display progress at each replicate.

   Outputs:
       M: Vector of module assignments (length n).
       Q: Value of maximized modularity.

   See also:
       MUMAP, LOYVAIN.

```



from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import abct
import numpy as np

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def degree(
    W: ArrayLike,
    type: Literal["first", "second", "residual"] = "first"
) -> np.ndarray:

    W = np.asarray(W)
    match type:
        case "first":
            return np.sum(W, axis=1)
        case "second":
            return np.sum(W**2, axis=1)
        case "residual":
            W_residual = abct.residualn(W, "rankone")
            return np.sum(W_residual, axis=1)


degree.__doc__ = files("abct").joinpath("docstrings", "doc_degree.py").read_text().replace("```", "")



import warnings

import numpy as np

def step4_run(Args, W, M, My=None):
    # Loyvain main algorithm

    # Unpack arguments
    k = Args.k
    n = len(M)

    MM = np.zeros((k, n))
    MM[M, np.arange(n)] = 1  # two-dimensional representation
    N = np.sum(MM, axis=1, keepdims=True)  # number of nodes in module
    match Args.method:
        case "loyvain":
            if Args.similarity == "network":
                Smn = MM @ W  # degree of module to node
            else:
                X = Args.X
                G = MM @ X  # cluster centroid
                Smn = G @ X.T  # dot of centroid with node
            Cii = np.diag(Smn @ MM.T)[:, np.newaxis]  # within-module weight sum
            if Args.objective == "spectral":
                S = np.sum(Smn, axis=0, keepdims=True)  # degree of node
                D = np.sum(Smn, axis=1, keepdims=True)  # degree of module
            Wii = Args.Wii[np.newaxis, :]  # within-node weight sum

        case "coloyvain":
            ny = len(My)
            MMy = np.zeros((k, ny))
            MMy[My, np.arange(ny)] = 1  # two-dimensional representation
            Ny = np.sum(MMy, axis=1, keepdims=True)  # number of nodes in module
            Smn = MMy @ W.T  # strength node to module of Wxy
            Cii = np.diag(MM @ Smn.T)[:, np.newaxis]  # within-module weight sum
            if Args.objective == "cospectral":
                S = np.sum(W, axis=1, keepdims=True)
                D = np.sum(MM @ W, axis=1, keepdims=True)
                Dy = np.sum(MMy @ W.T, axis=1, keepdims=True)

    match Args.objective:
        case "kmeans":      Cii_nrm = Cii / N
        case "spectral":    Cii_nrm = Cii / D
        case "cokmeans":    Cii_nrm = Cii / np.sqrt(N * Ny)
        case "cospectral":  Cii_nrm = Cii / np.sqrt(D * Dy)

    if (k == 1) or (k == n):
        Args.maxiter = 0  # skip loop if trivial partition

    for v in range(Args.maxiter):
        max_delta_Q = 0  # maximal increase over all batches
        idx = np.random.permutation(n)
        pts = np.round(np.linspace(0, n, Args.numbatches + 1)).astype(int)
        Batches = [np.array(idx[pts[i]:pts[i+1]]) for i in range(len(pts) - 1)]
        for u in range(Args.numbatches):
            U = Batches[u]  # indices of nodes in batch
            MU = M[U]  # module assignments of nodes in batch

            # Calculate change in modularity
            with np.errstate(divide="ignore", invalid="ignore"):
                match Args.objective:
                    case "kmeans":
                        delta_QU = (((2 * Smn[:, U] + Wii[:, U]) - Cii_nrm) / (N + 1) -
                                ((2 * Smn[MU, U] - Wii[:, U]) - Cii_nrm[MU].T) / (N[MU].T - 1))
                    case "spectral":
                        delta_QU = (((2 * Smn[:, U] + Wii[:, U]) - Cii_nrm * S[:, U]) / (D + S[:, U]) -
                                    ((2 * Smn[MU, U] - Wii[:, U]) - Cii_nrm[MU].T * S[:, U]) / (D[MU].T - S[:, U]))
                    case "cokmeans":
                        delta_QU = ((Cii + Smn[:, U]) / np.sqrt((N + 1) * Ny) - Cii_nrm + 
                                    (Cii[MU].T - Smn[MU, U]) / np.sqrt((N[MU].T - 1) * Ny[MU].T) - Cii_nrm[MU].T)
                    case "cospectral":
                        delta_QU = ((Cii + Smn[:, U]) / np.sqrt((D + S[:, U]) * Dy) - Cii_nrm + 
                                    (Cii[MU].T - Smn[MU, U]) / np.sqrt((D[MU].T - S[:, U]) * Dy[MU].T) - Cii_nrm[MU].T)

            delta_QU[:, (N[MU] == 1).ravel()] = -np.inf  # no change allowed if one-node cluster
            delta_QU[MU, np.arange(len(U))] = 0  # no change if node stays in own module

            # Update if improvements
            max_delta_QU = np.max(delta_QU, axis=0)
            MU_new = np.argmax(delta_QU, axis=0)
            if np.max(max_delta_QU) > Args.tolerance:
                max_delta_Q = np.maximum(max_delta_Q, np.max(max_delta_QU))

                IU = np.where(MU != MU_new)[0]  # batch indices of nodes to be switched
                I = U[IU]  # actual indices of nodes to be switched
                MI_new = MU_new[IU]  # new module assignments

                # get delta modules and ensure non-empty modules
                n_i = len(I)
                while True:
                    MMI_new = np.zeros((k, n_i))
                    MMI_new[MI_new, np.arange(n_i)] = 1
                    delta_MMI = MMI_new - MM[:, I]
                    N_new = N + np.sum(delta_MMI, axis=1, keepdims=True)
                    if np.all(N_new):
                        break
                    else:
                        E = np.where(N_new == 0)[0]  # empty modules
                        k_e = len(E)  # number of empty modules
                        MI_new[np.random.choice(n_i, k_e, replace=False)] = E

                # Update all relevant variables
                N = N_new
                M[I] = MI_new
                MM[:, I] = MMI_new

                match Args.method:
                    case "loyvain":
                        # Update G and Smn
                        if Args.similarity == "network":
                            delta_Smn = delta_MMI @ W[I]
                        else:
                            delta_G = delta_MMI @ X[I]  # change in centroid
                            G = G + delta_G  # update centroids
                            delta_Smn = delta_G @ X.T  # change in degree of module to node
                        Smn = Smn + delta_Smn  # update degree of module to node
                        Cii = np.diag(Smn @ MM.T)[:, np.newaxis]  # within-module weight sum
                    case "coloyvain":
                        Cii = np.diag(MM @ Smn.T)[:, np.newaxis]  # within-module weight sum
                        if Args.objective == "cospectral":
                            delta_Smn = delta_MMI @ W[I]

                if Args.objective in ["spectral", "cospectral"]:
                    D = D + np.sum(delta_Smn, axis=1, keepdims=True)

                match Args.objective:
                    case "kmeans":      Cii_nrm = Cii / N
                    case "spectral":    Cii_nrm = Cii / D
                    case "cokmeans":    Cii_nrm = Cii / np.sqrt(N * Ny)
                    case "cospectral":  Cii_nrm = Cii / np.sqrt(D * Dy)

        if max_delta_Q < Args.tolerance:
            break

        if (Args.display == "iteration") and (Args.method == "loyvain"):
            print(f"Replicate: {Args.replicate_i:4d}.    Iteration: {v:4d}.    Largest Δ: {max_delta_Q:4.4f}")
        if v == Args.maxiter:
            warnings.warn(f"Algorithm did not converge after {v} iterations.")

    # Return objective
    Q = np.sum(Cii_nrm)

    return M, Q, Cii_nrm



import numpy as np
from scipy import sparse
from abct import muma

def step3_init(Args):
    # m-umap output initialization

    A = Args.A
    M = Args.M
    Am = Args.Am
    k = Args.k
    match Args.start:
        case "custom":
            U = Args.U
            U = U / np.linalg.norm(U, axis=1, keepdims=True)
        case "spectral_nn":                         # spectral on knn matrix
            _, U = sparse.linalg.eigs(A, Args.d+1)
            U = U[:, 1:]
            U = U / np.linalg.norm(U, axis=1, keepdims=True)
        case "spectral":                            # spectral on modules
            Amm = M.T @ Am
            Kmm = np.sum(Amm, axis=1, keepdims=True)
            Bmm = Amm - Kmm * Kmm.T / np.sum(Amm)
            _, Um = sparse.linalg.eigs(Bmm, Args.d)
            Um = Um / np.linalg.norm(Um, axis=1, keepdims=True)
            U = Um[Args.partition]
        case "greedy":                              # spherical maximin
            Amm = M.T @ Am                          # module connectivity
            np.fill_diagonal(Amm, np.nan)           # ignore self-connections
            Kmm_ = np.zeros(k)                      # degree to placed modules
            Um = np.zeros((k, Args.d))              # module locations
            Vm = muma.fsphere(k)                    # Fibonacci sphere
            ux = np.argmax(np.nansum(Amm, axis=1))  # initial module index
            vx = np.argmax(np.nansum(Vm @ Vm.T, axis=1)) # initial location index
            for i in range(k):
                Um[ux] = Vm[vx]                     # assign location
                Vm[vx] = np.nan                     # remove point from consideration
                if i == k-1:
                    break
                Kmm_ = Kmm_ + Amm[:, ux]            # add module connectivity (with self-nan's)
                ux = np.nanargmin(Kmm_)             # least connected module (nan's in Kmm mask set modules)
                vx = np.nanargmin(Vm @ np.mean(Um, axis=0, keepdims=True).T)  # furthest location (nan's in Vm mask used locations)
            U = Um[Args.partition]

    return U



import abct
import numpy as np

def step1_proc_loyvain(Args):
    # Loyvain arguments processing

    Args.n, Args.p = Args.X.shape
    if Args.similarity == "network":
        Args.W = Args.X
        Args.X = np.array(None)
    else:
        Args.W = np.array(None)

    # Process custom initial module assignment
    if isinstance(Args.start, np.ndarray):
        Args.M0 = np.ravel(Args.start).astype(int)
        if Args.k == 0:
            Args.k = np.max(Args.M0) + 1
        Args.start = "custom"

    # Global residualization for kmodularity
    if Args.objective == "kmodularity":
        if Args.similarity == "network":
            Args.W = Args.W * (Args.n / Args.k) / np.sum(np.abs(Args.W))
            Args.W = abct.residualn(Args.W, "degree")
        else:
            Args.X = abct.residualn(Args.X, "global")
        Args.objective = "kmeans"

    # Center to mean 0 for covariance and correlation
    if Args.similarity in ["cov", "corr"]:
        Args.X = Args.X - np.mean(Args.X, axis=1, keepdims=True)

    # Normalize to norm 1 for cosine and correlation
    if Args.similarity in ["cosim", "corr"]:
        Args.X = Args.X / np.linalg.norm(Args.X, axis=1, keepdims=True)
    elif Args.similarity in ["dot", "cov"]:
        Args.X = Args.X / np.sqrt(Args.p)

    # Compute self-connection weights
    if Args.similarity == "network":
        Args.Wii = np.diag(Args.W)
    else:
        Args.Wii = np.sum(Args.X**2, axis=1)

    # Precompute kmeans++ variables
    Args.Dist, Args.normX = [0], [0]
    if Args.start in ["greedy", "balanced"]:
        if Args.similarity == "network":
            Args.Dist = Args.W / np.linalg.norm(Args.W, axis=1, keepdims=True)
            Args.Dist = 1 - Args.Dist @ Args.Dist.T
        else:
            Args.normX = np.linalg.norm(Args.X, axis=1, keepdims=True)

    return Args



"""
Loyvain subfunctions.
"""

from abct.loyv.step0_args import step0_args
from abct.loyv.step1_proc_loyvain import step1_proc_loyvain
from abct.loyv.step1_proc_coloyvain import step1_proc_coloyvain
from abct.loyv.step2_test import step2_test
from abct.loyv.step3_init import step3_init
from abct.loyv.step4_run import step4_run

__all__ = [
    "step0_args",
    "step1_proc_loyvain",
    "step1_proc_coloyvain",
    "step2_test",
    "step3_init",
    "step4_run",
]



import warnings
from typing import Tuple
from importlib.resources import files

import numpy as np
from abct import loyv

def coloyvain(*args, **kwargs) -> Tuple[np.ndarray, np.ndarray, float, np.ndarray]:

    ## Parse, process, and test arguments

    # Parse arguments
    Args = loyv.step0_args("coloyvain", *args, **kwargs)

    # Process initial arguments
    Args = loyv.step1_proc_coloyvain(Args)

    # Test arguments
    loyv.step2_test(Args.X, Args.W, Args.px, Args.k, Args)
    loyv.step2_test(Args.Y, Args.W, Args.py, Args.k, Args)

    ## Run algorithm

    R = -np.inf
    for i in range(Args.replicates):
        Args.replicate_i = i

        # initialize
        Mx0 = loyv.step3_init([], [], Args.DistX, Args.px, Args)
        My0 = loyv.step3_init([], [], Args.DistY, Args.py, Args)

        # get between-module correlations
        MMx0 = np.zeros((Args.k, Args.px))
        MMx0[Mx0, np.arange(Args.px)] = 1
        MMy0 = np.zeros((Args.k, Args.py))
        MMy0[My0, np.arange(Args.py)] = 1

        match Args.objective:
            case "cokmeans":
                Ox = np.ones((Args.px, 1))
                Oy = np.ones((Args.py, 1))
            case "cospectral":
                Ox = np.sum(Args.W, axis=1, keepdims=True)
                Oy = np.sum(Args.W, axis=0, keepdims=True).T

        C0_nrm = (MMx0 @ Args.W @ MMy0.T) / np.sqrt((MMx0 @ Ox) @ (MMy0 @ Oy).T)

        # align modules
        Mx1 = np.zeros_like(Mx0)
        My1 = np.zeros_like(My0)
        for h in range(Args.k):
            ix, iy = np.unravel_index(np.nanargmax(C0_nrm), C0_nrm.shape)
            Mx1[Mx0 == ix] = h
            My1[My0 == iy] = h
            C0_nrm[ix, :] = np.nan
            C0_nrm[:, iy] = np.nan

        # fixed point iteration until convergence
        for v in range(Args.maxiter):
            My0 = My1.copy()
            Mx1,  _,      _ = loyv.step4_run(Args, Args.W, Mx1, My1)  # optimize Mx
            My1, R1, R1_all = loyv.step4_run(Args, Args.W.T, My1, Mx1)  # optimize My
            if np.array_equal(My0, My1):  # if identical, neither Mx1 nor My1 will change
                break
            if Args.display == "iteration":
                print(f"Replicate: {Args.replicate_i:4d}.    Iteration: {v:4d}.    Objective: {R1:4.4f}.")
            if v == Args.maxiter - 1:
                warnings.warn(f"Algorithm did not converge after {v + 1} iterations.")

        # check if replicate has improved on previous result
        if (R1 - R) > Args.tolerance:  # test for increase
            if Args.display in ["replicate", "iteration"]:
                print(f"Replicate: {i:4d}.    Objective: {R1:4.4f}.    Δ: {R1 - R:4.4f}.")
            R = R1
            Mx = Mx1
            My = My1
            R_all = R1_all

    return Mx, My, R, R_all.ravel()


coloyvain.__doc__ = files("abct").joinpath("docstrings", "doc_coloyvain.py").read_text().replace("```", "")



```
 CANONCOV Canonical covariance analysis (aka partial least squares)
          Canonical correlation analysis

   A, B, U, V, R = canoncov(X, Y, k)
   A, B, U, V, R = canoncov(X, Y, k, type)
   A, B, U, V, R = canoncov(X, Y, k, type, resid)
   A, B, U, V, R = canoncov(X, Y, k, type, resid, corr)
   A, B, U, V, R = canoncov(X, Y, k, type, resid, corr, Name=Value)

   Inputs:
       X: Data matrix of size n x p, where
          n is the number of data points and
          p is the number of features.

       Y: Data matrix of size n x q, where
          n is the number of data points and
          q is the number of features.

       k: Number of canonical components (positive integer).

       type: Weighted or binary canonical analysis.
           "weighted": Weighted canonical analysis (default).
           "binary": Binary canonical analysis.

       resid: Global residualization (logical scalar).
           0: No global residualization.
           1: Global residualization via degree correction (default).

       corr: Canonical correlation analysis (logical scalar).
           0: Canonical covariance analysis (default).
           1: Canonical correlation analysis.

       Name=Value Arguments
           (binary canonical analysis only):
           See LOYVAIN for all Name=Value options.

   Outputs:
       A: Canonical coefficients of X (size p x k).
       B: Canonical coefficients of Y (size q x k).
       U: Canonical components of X (size n x k).
       V: Canonical components of Y (size n x k).
       R: Canonical covariances or correlations (size k x k).
          If type is "weighted", R denotes the actual covariances or
          correlations. If type is "binary", R denotes the
          normalized covariances or correlations.

   Methodological notes:
       Weighted canonical correlation or covariance analysis is computed
       via singular value decomposition of cross-covariance matrix.

       Binary canonical covariance analysis is computed via co-Loyvain
       k-means clustering of cross-covariance matrix. This analysis
       produces binary orthogonal canonical coefficients.

       Binary canonical covariance analysis is computed via co-Loyvain
       k-means clustering of _whitened_ cross-covariance matrix. This
       analysis produces binary orthogonal canonical coefficients for
       the whitened matrix. However, the output coefficients after
       dewhitening will, in general, not be binary.

       Global residualization is implemented via generalized degree
       correction, and converts k-means co-clustering into k-modularity
       co-maximization.

   See also:
       COLOYVAIN, LOYVAIN, RESIDUALN.

```



import torch
import pymanopt
import numpy as np
from scipy import sparse
import warnings

def step4_run(U, Args):
    # m-umap main algorithm

    ## Initialize GPU arrays
    device = "cuda" if Args.gpu else "cpu"

    k = Args.k
    gamma = Args.gamma
    alpha = Args.alpha
    beta = Args.beta
    A = Args.A
    M = Args.M
    Am = Args.Am

    ## Precompute gradient matrices

    # Normalized degree column vector
    K_nrm = np.sqrt(gamma / A.sum()) * A.sum(1)[:, np.newaxis]

    # Modules and normalized modules
    N = M.sum(0, keepdims=True)
    M_nrm = M / N

    # Module adjacency and modularity matrices
    g = gamma * A.mean()     # mean(K)^2 / sum(K)
    # Bm == ((A - g) .* (~(M * M'))) * M
    #    == ((A - g) * M - ((A - g) .* (M * M')) * M
    #    -> then simplify (A - g) .* (M * M')
    Bm = (Am - M * Am) - (g * N - g * (M * N))

    Ic = [np.array([])] * k
    Bc = [np.array([])] * k
    Ac = [sparse.csr_array([])] * k
    Kc_nrm = [np.array([])] * k
    for i in range(k):
        I = np.where(Args.partition == i)[0]
        Ic[i] = I
        if Args.cache:
            Bc[i] = A[np.ix_(I, I)] - (K_nrm[I] * K_nrm[I].T)
        else:
            Ac[i] = A[np.ix_(I, I)]
            Kc_nrm[i] = K_nrm[I]

    ## Initialize Torch tensors

    U = torch.as_tensor(U, device=device).contiguous().requires_grad_(True)
    M_nrm = torch.as_tensor(M_nrm, device=device)
    Bm = torch.as_tensor(Bm, device=device)
    alpha = torch.as_tensor(alpha, device=device)
    beta = torch.as_tensor(beta, device=device)
    for i in range(k):
        Ic[i] = torch.as_tensor(Ic[i], device=device)
        Bc[i] = torch.as_tensor(Bc[i], device=device)
        with warnings.catch_warnings(action="ignore"):
            Ac[i] = torch.sparse_csr_tensor(Ac[i].indptr, Ac[i].indices, Ac[i].data, device=device)
        Kc_nrm[i] = torch.as_tensor(Kc_nrm[i], device=device)

    ## Run solvers

    match Args.solver:
        case "adam":
            vb = Args.verbose
            fp = {
                "head": lambda: print("%5s %24s %12s" % ("iter", "cost val", "grad. norm")) if vb else None,
                "iter": lambda t, cost, grad_norm: print("%5d %+.16e %12e" % (t, cost, grad_norm)) if vb else None,
                "stop_cost": lambda: print("Cost tolerance reached; tolerance = %g." % Args.tolerance) if vb else None,
                "stop_grad": lambda: print("Gradient norm tolerance reached; tolerance = %g." % Args.tolerance) if vb else None,
                "stop_iter": lambda: print("Max iter exceeded; maxiter = %g." % Args.maxiter) if vb else None
            }
            fp["head"]()

            optimizer = torch.optim.Adam([U], lr=Args.learnrate)
            CostHistory = np.full(Args.maxiter, np.nan)
            for t in range(Args.maxiter):
                cost = fx_cost(U, Ic, Bc, Ac, Kc_nrm, M_nrm, Bm, alpha, beta)
                optimizer.zero_grad()
                cost.backward()
                with torch.no_grad():  # Get Riemannian gradient
                    U_dot_EGrad = (U * U.grad).sum(dim=1, keepdim=True)
                    U.grad -= U * U_dot_EGrad
                optimizer.step()

                grad_norm = U.grad.norm().detach().cpu().numpy()
                with torch.no_grad():
                    U /= U.norm(dim=1, keepdim=True)

                cval = cost.detach().cpu().numpy()
                CostHistory[t] = cval
                fp["iter"](t, cval, grad_norm)
                if t and (abs(cval - CostHistory[t-1]) < Args.tolerance):
                    fp["stop_cost"]()
                    break
                elif grad_norm < Args.tolerance:
                    fp["stop_grad"]()
                    break
                elif t == Args.maxiter - 1:
                    fp["stop_iter"]()
                    break

            U = U.detach().cpu().numpy()
            CostHistory = CostHistory[~np.isnan(CostHistory)]

        case "trustregions":
            # Create the problem structure.
            manifold = pymanopt.manifolds.Oblique(Args.d, Args.n)   # transposed to normalize rows
            fx_ucost = lambda U: fx_cost(U.T, Ic, Bc, Ac, Kc_nrm, M_nrm, Bm, alpha, beta)
            problem = pymanopt.Problem(manifold, cost=pymanopt.function.pytorch(manifold)(fx_ucost))
            optimizer = pymanopt.optimizers.TrustRegions(max_time=np.inf)
            result = optimizer.run(problem)

            U = result.point.T                                      # transposed back
            CostHistory = result.cost

    return U, CostHistory

def fx_cost(U, Ic, Bc, Ac, Kc_nrm, M_nrm, Bm, alpha, beta):

    k = len(Ic)

    ## Compute mean-field between-module cost
    # UUm == ((U * U') .* (~(M * M'))) * Mn
    UUm = U @ (U.T @ M_nrm)
    for i in range(k):
        I = Ic[i]
        UUm[I, i] = 0          # exclude self-modules

    Dm = 2 * (1 - UUm)
    Numm = beta * alpha * (Dm ** (beta - 1))
    if beta >= 1:                # fast update
        Denm = 1 + Numm * Dm / beta
    else:                        # avoid NaN
        Denm =  1 + alpha * (Dm ** beta)

    Cost = - torch.sum(Bm / Denm)

    ## Compute full within-module cost and gradient
    for i in range(k):
        if len(Bc[i]):           # Args.cache is True
            Bi = Bc[i]
        else:
            Bi = Ac[i].to_dense() - (Kc_nrm[i] * Kc_nrm[i].T)

        I = Ic[i]
        Ui = U[I]
        ni = len(Ui)   # number of nodes in module i
        Di = 2 * (1 - (Ui @ Ui.T))
        Numi = beta * alpha * (Di ** (beta - 1))
        Numi.fill_diagonal_(0)
        if beta >= 1:            # fast update
            Deni = 1 + Numi * Di / beta
        else:                    # avoid NaN
            Deni =  1 + alpha * (Di ** beta)

        Cost -= torch.sum(Bi / Deni)

    return Cost

def fx_cost_full(U, B, alpha, beta):
    ## Compare full cost and gradient
    D = 2 * (1 - (U @ U.T))
    Num = beta * alpha * (D ** (beta - 1))
    Num.fill_diagonal_(0)
    Den1 =   1 + alpha * (D ** beta)
    Cost =  - torch.sum(B / Den1)

    return Cost




```
 LOYVAIN K-modularity, k-means, or spectral clustering

   M, Q = loyvain(W, k)
   M, Q = loyvain(W, k, objective)
   M, Q = loyvain(X, k, objective, similarity)
   M, Q = loyvain(_, objective, similarity, Name=Value)

   Inputs:
       W:  Network matrix of size n x n.
       OR
       X:  Data matrix of size n x p, where
           n is the number of data points and
           p is the number of features.

       k: Number of modules (positive integer or 0).
           Set to 0 to infer number from initial module assignment.

       objective: Clustering objective.
           "kmodularity": K-modularity (default).
           "kmeans": K-means clustering objective.
           "spectral": Spectral clustering objective (normalized cut).

       similarity: Type of similarity.
         The first option assumes that the first input is a network matrix.
           "network": Network connectivity (default).
               W is a symmetric network matrix. The network must
               be non-negative for k-modularity and spectral
               objectives. No additional similarity is computed.
         The other options assume that the first input is X, a data matrix.
           "corr": Pearson correlation coefficient.
               A magnitude-normalized dot product of mean-centered vectors.
           "cosim": Cosine similarity.
               A normalized dot product.
           "cov":  Covariance.
               A dot product of mean-centered vectors.
           "dot": Dot product.
               A sum of an elementwise vector product.

       Name=Value Arguments:

           start=Initial module assignments.
               "greedy": Maximin (greedy kmeans++) initialization (default).
               "balanced": Standard kmeans++ initialization.
               "random": Uniformly random initialization.
               Numeric vector: Initial module assignment vector of length n.

           numbatches=Number of batches.
               Positive integer (default is 10).

           maxiter=Maximum number of algorithm iterations.
               Positive integer (default is 1000).

           replicates=Number of replicates.
               Positive integer (default is 10).

           tolerance=Convergence tolerance.
               Positive scalar (default is 1e-10).

           display=Display progress.
               "none": no display (default).
               "replicate": display progress at each replicate.
               "iteration": display progress at each iteration.

   Outputs:
       M: Vector of module assignments (length n).
       Q: Value of maximized objective.

   Methodological notes:
       Loyvain is a unification of:
       Lloyd's algorithm for k-means clustering and
       Louvain algorithm for modularity maximization.

       K-modularity maximization is exactly equivalent to normalized
       modularity maximization and approximately equivalent to k-means
       clustering with global residualization. Global residualization is
       implemented as degree correction for network matrices and
       global-signal regression for data matrices.

       For "network" similarity, k-modularity is rescaled by:
           (average module size) / (absolute sum of all weights)
       This rescaling aligns k-modularity within the range of the
       modularity, but has no effect on the optimization algorithm.

       The Loyvain algorithm is not guaranteed to converge if
       all swaps are accepted at each iteration (NumBatches = 1).
       Therefore, it is generally a good idea to set NumBatches > 1.

   See also:
       COLOYVAIN, CANONCOV, KNEICOMP, RESIDUALN.

```



import warnings
from typing import Literal, Tuple
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import abct
import numpy as np
from scipy import linalg, sparse

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def canoncov(
    X: ArrayLike,
    Y: ArrayLike,
    k: int,
    type: Literal["weighted", "binary"] = "weighted",
    resid: bool = True,
    corr: bool = False,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

    X = np.asarray(X)
    Y = np.asarray(Y)

    # Basic checks
    n, p = X.shape
    n_, q = Y.shape
    if n != n_:
        raise ValueError("X and Y must have the same number of data points.")
    if k > min(p, q):
        raise ValueError("k must not exceed number of features in X or Y.")

    # Initial processing
    if type == "weighted":
        if kwargs:
            warnings.warn("Ignoring Name=Value arguments for weighted analysis.")

    # Global residualization or centering
    if resid:  # Degree correction automatically centers data
        X = abct.residualn(X, "degree")
        Y = abct.residualn(Y, "degree")
    else:
        X -= np.mean(X, axis=0, keepdims=True)
        Y -= np.mean(Y, axis=0, keepdims=True)

    # Set up problem
    if corr:
        def truncate(U, S, Vt):
            tol = np.finfo(S.dtype).eps ** (0.25) # matches MATLAB default
            ix = np.where((1 - np.cumsum(S**2)/np.sum(S**2)) > tol)[0]
            return U[:, ix], S[ix], Vt[ix, :]
        Ux, Sx, Vx = truncate(linalg.svd(X, full_matrices=False))
        Uy, Sy, Vy = truncate(linalg.svd(Y, full_matrices=False))
        Vx = Vx.T
        Vy = Vy.T
        Z = Vx @ Ux.T @ Uy @ Vy.T
    else:
        Z = X.T @ Y

    # Solve problem
    if type == "weighted":
        A, R, B = sparse.linalg.svds(Z, k=k)
        B = B.T
    else:
        Mx, My, _, R = abct.coloyvain(Z, k, "kmeans", "network", numbatches=min(32, min(p, q)), **kwargs)
    ix = np.argsort(R)[::-1]
    R = R[ix]
    if type == "weighted":
        A = A[:, ix]
        B = B[:, ix]
    else:
        A = np.zeros((p, k))
        B = np.zeros((q, k))
        for h in range(k):
            A[Mx == ix[h], h] = 1
            B[My == ix[h], h] = 1

    # Recover coefficients
    if corr:
        A = (Vx / Sx) @ Vx.T @ A
        B = (Vy / Sy) @ Vy.T @ B

    U = X @ A
    V = Y @ B

    return A, B, U, V, R


canoncov.__doc__ = files("abct").joinpath("docstrings", "doc_canoncov.py").read_text().replace("```", "")



```
 MUMAP m-umap low-dimensional embedding

   U = mumap(X);
   U, Partition, CostHistory = mumap(X, Name=Value);

   Inputs:
       W:  Network matrix of size n x n.

       OR

       X:  Data matrix of size n x p, where
           n is the number of data points and
           p is the number of features.

       Name=Value Arguments:
           d=Embedding dimension.
               Positive integer (default is 3).

           kappa=Number of nearest neighbors.
               Positive integer (default is 30).

           alpha=Inverse amplitude of long-range attraction.
               Positive scalar >= 1 (default is 1).
               Larger alpha implies weaker long-range attraction.

           beta=Slope of long-range attraction.
               Positive scalar <= 1 (default is 1).
               Larger beta implies faster decay of attraction.

           gamma=Modularity resolution parameter.
               Positive scalar (default is 1).

           similarity=Type of similarity.
               "network": Network connectivity (default).
               "corr": Pearson correlation coefficient.
               "cosim": Cosine similarity.

           method=Method of nearest-neighbor search.
               "direct": Direct computation of similarity matrix.
               "indirect": knnsearch (MATLAB) or pynndescent (Python).

           replicates=Number of modularity replicates.
               Positive integer (default is 10).

           finaltune=Modularity final tuning.
               Logical scalar (default is true).

           partition=Module partition.
               Integer vector: module partition of length n (default is ).

           start=Initial embedding method.
               "greedy": Spherical maximin initialization (default).
               "spectral": Spectral initialization of module (k x k) matrix.
               "spectral_nn": Spectral initialization of full (n x n) matrix.
               Numeric matrix: Initial embedding of size n x d, where:
                   n is the number of data points and
                   d is the embedding dimension.

           solver=Optimization solver.
               "trustregions": Manopt trust-regions method (default for MATLAB).
               "adam": Adaptive Moment Estimation optimizer (default for Python).

           maxiter=Maximum number of iterations.
               Positive integer (default is 10000).

           learnrate=Optimizer learning rate.
               Positive scalar (default is 0.001).

           tolerance=Solution tolerance.
               Positive scalar (default is 1e-6).

           gpu=Use GPU.
               Logical (default is false).

           cache=Cache gradient matrices.
               Logical (default is false).

           verbose=Verbose output.
               Logical (default is true).

   Outputs:
       U:  Embedding matrix of size n x d.
       Partition: Module partition of length n.
       CostHistory: Cost history of optimization.

   Methodological notes:
       m-umap is a first-order approximation of the true parametric
       loss of UMAP, with spherical constraints. It is simultaneously
       equivalent to the modularity, and to spring layout methods with
       Cauchy components and spherical constraints.

   Dependencies:
       MATLAB: 
           Statistics and Machine Learning Toolbox (if method="indirect")
           Deep Learning Toolbox (if solver="adam")
           Manopt (if solver="trustregions")
       Python: 
           igraph
           PyTorch
           PyNNDescent (if method="indirect")
           PyManopt (if solver="trustregions") * warning: slow *

```



from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import numpy as np
from scipy import sparse


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def residualn(
    X: ArrayLike,
    type: Literal["degree", "global", "rankone"] = "degree"
) -> np.ndarray:

    X = np.asarray(X)

    match type:
        # Degree correction
        case "degree":
            if not np.all(X >= 0):
                raise ValueError("Invalid degree correction: Matrix must be non-negative.")
            So = np.sum(X, axis=1)
            Si = np.sum(X, axis=0)
            return X - np.outer(So, Si) / np.sum(So)

        # Global signal regression
        case "global":
            G = np.mean(X, axis=0, keepdims=True)
            return X - (X @ G.T) @ G / (G @ G.T)

        # Subtraction of rank-one approximation
        case "rankone":
            U, S, VT = sparse.linalg.svds(X, k=1)
            return X - U @ np.diag(S) @ VT


residualn.__doc__ = files("abct").joinpath("docstrings", "doc_residualn.py").read_text().replace("```", "")



"""
m-umap subfunctions.
"""

from abct.muma.fsphere import fsphere
from abct.muma.rotation import rotation
from abct.muma.projection import projection
from abct.muma.step0_args import step0_args
from abct.muma.step1_proc import step1_proc
from abct.muma.step2_test import step2_test
from abct.muma.step3_init import step3_init
from abct.muma.step4_run import step4_run

__all__ = [
    "fsphere",
    "rotation",
    "projection",
    "step0_args",
    "step1_proc",
    "step2_test",
    "step3_init",
    "step4_run",
]



import warnings

import numpy as np

def step2_test(X, W, n, k, Args):
    # Loyvain arguments tests

    if k < 1:
        raise ValueError("Specify number of modules or starting module assignment.")
    if k > n:
        raise ValueError("Number of modules must not exceed number of nodes or data points.")
    if Args.numbatches > n:
        raise ValueError("Number of batches must not exceed number of nodes or data points.")
    if not (np.array_equal(X, None) or np.all(np.isfinite(X))):
        raise ValueError("Data matrix has non-finite elements after processing.")

    # Test non-negativity for spectral clustering
    if Args.objective in ["spectral", "cospectral"]:
        ending = 'non-negative for "spectral" objective.'
        if Args.similarity == "network":
            if not np.all(W >= 0):
                raise ValueError("Network matrix must be " + ending)
        elif X.size < 1e6:
            if not np.all(X @ X.T >= 0):
                raise ValueError("Similarity matrix must be " + ending)
        else:
            warnings.warn(
                "Not checking similarity matrix for negative values because "
                "of large data size. Ensure that similarity matrix is " + ending
            )

    if Args.method == "loyvain":
        # Test symmetry
        if not ((W.shape[0] == W.shape[1]) and np.allclose(W, W.T)):
            raise ValueError('Network matrix must be symmetric or similarity must not be "network".')

        # Test initialization
        if Args.start == "custom":
            if not ((len(Args.M0) == n) and np.array_equal(np.unique(Args.M0), np.arange(k))):
                raise ValueError(f"Initial module assignment must have length {n} and contain integers {0} to {k-1}.")



from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import numpy as np

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def dispersion(
    W: ArrayLike,
    type: Literal["coefvar2", "kpartcoef"] = "coefvar2",
    M: ArrayLike = [],
) -> np.ndarray:

    W = np.asarray(W)
    M = np.asarray(M)

    match type:
        case "coefvar2":
            D = np.var(W, axis=1) / (np.mean(W, axis=1) ** 2)
        case "kpartcoef":
            # Basic checks
            n, n_ = W.shape
            n__ = M.size
            if n != n_ or n != n__:
                raise ValueError("W must be a square matrix and M must have the same length as W.")

            k = np.max(M) + 1
            MM = np.zeros((n, k))
            MM[np.arange(n), M] = 1
            kSnm = (W @ MM) / MM.sum(0, keepdims=True)
            P = kSnm / kSnm.sum(1, keepdims=True)
            D = 1 - (P**2).sum(1)

    return D


dispersion.__doc__ = files("abct").joinpath("docstrings", "doc_dispersion.py").read_text().replace("```", "")



from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import numpy as np
import igraph as ig
from scipy import sparse

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def leiden(
    W,
    gamma: float = 1,
    start: ArrayLike = None,
    replicates: int = 10,
    finaltune: bool = False,
    tolerance: float = 1e-10,
    display: Literal["none", "replicate"] = "none",
):
    W = sparse.csr_matrix(W) if sparse.issparse(W) else np.asarray(W)
    G = ig.Graph.Weighted_Adjacency(W, mode="undirected", attr="weight")
    Q = -np.inf
    for i in range(replicates):
        replicate_i = i

        # run algorithm
        C = G.community_leiden(
            objective_function="modularity",
            weights="weight",
            resolution=gamma,
            initial_membership=start,
        )
        if finaltune:
            C = G.community_leiden(
                objective_function="modularity",
                weights="weight",
                resolution=gamma,
                initial_membership=C.membership,
            )

        # test for increase
        if (C.modularity - Q) > tolerance:
            if display == "replicate":
                print("Replicate: %4d.    Objective: %4.4f.    Δ: %4.4f." % (i, C.modularity, C.modularity - Q))
            
            Q = C.modularity
            M = C.membership

    return np.asarray(M), Q

leiden.__doc__ = files("abct").joinpath("docstrings", "doc_louvains.py").read_text().replace("```", "")



import abct
import numpy as np

def step1_proc(Args):
    # m-umap arguments processing

    # Process custom initial embedding
    if isinstance(Args.start, np.ndarray):
        Args.U = Args.start
        Args.start = "custom"

    # Generate a nearest-neighbor matrix
    Args.A = abct.kneighbor(Args.X,
        "nearest",
        Args.kappa,
        Args.similarity,
        Args.method,
    )

    # Module structure
    if Args.partition is None:
        Args.partition = abct.leiden(Args.A,
            gamma=Args.gamma,
            replicates=Args.replicates,
            finaltune=Args.finaltune,
            display="replicate" if Args.verbose else "none",
        )[0]
    else:
        Args.partition = np.unique_inverse(Args.partition).inverse_indices

    Args.n = len(Args.X)
    Args.k = np.max(Args.partition) + 1
    Args.M = np.zeros((Args.n, Args.k))
    Args.M[np.arange(Args.n), Args.partition] = 1
    Args.Am = Args.A @ Args.M

    return Args



from numpy.typing import ArrayLike

import numpy as np

def projection(U3: ArrayLike) -> np.ndarray:
    # Mercator projection

    X, Y, Z = np.array(U3).T
    U2 = np.column_stack((np.arctan2(Y, X), np.log((1 + Z)/(1 - Z))/2))

    return U2



from typing import Literal
from numpy.typing import ArrayLike
from scipy.sparse import sparray
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import numpy as np
import pynndescent
from scipy import sparse

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def kneighbor(
    X: ArrayLike | sparray,
    type: Literal["common", "nearest"] = "common",
    kappa: float = 10,
    similarity: Literal["network", "corr", "cosim"] = "network",
    method: Literal["direct", "indirect"] = "direct",
    **kwargs,
) -> sparse.csr_array:

    n = len(X)

    if kappa < 1:
        kappa = np.clip(np.round(n * kappa), 1, n-1)
    else:
        assert kappa < n, "kappa must be less than number of nodes or data points."
        assert np.isclose(kappa, np.round(kappa)), "kappa > 1 must be an integer."
    kappa = int(kappa)

    Row = np.tile(np.arange(n)[:, np.newaxis], (1, kappa+1))
    if similarity == "network":
        W = X.copy()
        assert W.shape[0] == W.shape[1], "Network matrix must be square."
        np.fill_diagonal(W, np.inf)
        Col = np.argpartition(W, -(kappa+1), axis=1)[:, -(kappa+1):].ravel()
    else:
        match method:
            case "direct":
                # Center to mean 0 for correlation
                if similarity == "corr":
                    X = X - np.mean(X, axis=1, keepdims=True)
                # Rescale to norm 1
                X = X / np.linalg.norm(X, axis=1, keepdims=True)

                # Compute similarity matrix in blocks of 1e8 elements
                # It follows that n * nb = 1e8, b = n / nb = n^2 / 1e8
                b = int(np.ceil(n**2 / 1e8))
                b = np.clip(b, 1, n)
                Ix = np.floor(np.linspace(0, n, b+1)).astype(int)
                Col = np.zeros((n, kappa+1), dtype=int)
                for i in range(b):
                    Ixi = np.arange(Ix[i], Ix[i+1])
                    Col[Ixi] = np.argpartition(X[Ixi, :] @ X.T, -(kappa+1), axis=1)[:, -(kappa+1):]

            case "indirect":
                match similarity:
                    case "corr": knnsim = "correlation"
                    case "cosim": knnsim = "cosine"
                knn_search_index = pynndescent.NNDescent(X, n_neighbors=kappa+1, metric=knnsim, **kwargs)
                Col, _ = knn_search_index.neighbor_graph

    A = sparse.csr_array((np.ones(n*(kappa+1)), (Row.ravel(), Col.ravel())), shape=(n, n))
    A.setdiag(0)

    match type:
        case "common":
            B = A @ A.T
        case "nearest":
            B = A.maximum(A.T)

    return B

kneighbor.__doc__ = files("abct").joinpath("docstrings", "doc_kneighbor.py").read_text().replace("```", "")



```
 DEGREE Degree of network matrix

   S = degree(W)
   S = degree(W, type)

   Inputs:
       W: Network matrix of size n x n.

       type: Degree type
           "first": (First) degree (default).
           "second": Second degree.
           "residual": Degree after global residualization.

   Outputs:
       S: Degree vector (length n).

   Methodological notes:
       The first degree is the sum of connection weights. The second
       degree is the sum of squared connection weights. Together, the
       first and second degrees are exactly or approximately equivalent to
       several measures of network communication and control.

       The residual degree is the degree after first-component removal and
       can be approximately equivalent to the primary co-activity gradient
       in functional MRI co-activity networks.

   See also:
       RESIDUALN, DISPERSION.

```



from typing import Tuple
from importlib.resources import files

import numpy as np
from . import loyv

def loyvain(*args, **kwargs) -> Tuple[np.ndarray, float]:

    ## Parse, process, and test arguments

    # Parse arguments
    Args = loyv.step0_args("loyvain", *args, **kwargs)

    # Process all other arguments
    Args = loyv.step1_proc_loyvain(Args)

    # Test arguments
    loyv.step2_test(Args.X, Args.W, Args.n, Args.k, Args)

    ## Run algorithm

    Q = -np.inf
    for i in range(Args.replicates):
        Args.replicate_i = i
        if Args.start == "custom":
            M0 = Args.M0
        else:
            # initialize
            M0 = loyv.step3_init(Args.X, Args.normX, Args.Dist, Args.n, Args)

        # run algorithm
        M1, Q1, _ = loyv.step4_run(Args, Args.W, M0)

        # test for increase
        if (Q1 - Q) > Args.tolerance:
            if Args.display in ["replicate", "iteration"]:
                print(
                    f"Replicate: {i:4d}.    Objective: {Q1:4.4f}.    Δ: {Q1 - Q:4.4f}."
                )
            Q = Q1
            M = M1

    return M, Q


loyvain.__doc__ = files("abct").joinpath("docstrings", "doc_loyvain.py").read_text().replace("```", "")



import numpy as np

def step2_test(Args):
    # m-umap arguments tests

    if Args.d >= Args.n:
        raise ValueError("Embedding dimension must be less than number of nodes or data points.")

    # Test partition
    if (len(Args.partition) != Args.n) or not np.array_equal(np.unique(Args.partition), np.arange(Args.k)):
        raise ValueError("Initial module partition must have length %d and contain integers 0 to %d." % (Args.n, Args.k - 1))

    # Test initial embedding
    if Args.start == "custom":
        if Args.U.shape != (Args.n, Args.d):
            raise ValueError("Initial embedding must have %d rows and %d columns." % (Args.n, Args.d))

    # Test initializations
    match Args.start:
        case "greedy":
            if Args.d != 3:
                raise ValueError("Embedding dimension must be 3 for ""greedy"" initialization.")
        case "spectral":
            if Args.d > Args.k:
                raise ValueError("Number of modules is too small for ""spectral"" initialization.")
        case "spectral_nn":
            if Args.d >= Args.n - 1:
                raise ValueError("Embedding dimension must be < n - 1 for ""spectral_nn"" initialization.")




```
 RESIDUALN Global residualization of network or timeseries data

   X1 = residualn(X)
   X1 = residualn(X, type)

   Inputs:
       X:  Network matrix of size n x n, or data matrix of size n x p.
           n is the number of nodes or data points and
           p is the number of features.

       type: Type of global residualization.
           "degree": Degree correction (default).
           "global": Global signal regression.
           "rankone": Subtraction of rank-one approximation.

   Outputs:
       X1: Residual network or timeseries matrix.

   See also:
       SHRINKAGE.

```



import abct
import numpy as np

def step1_proc_coloyvain(Args):
    # co-Loyvain arguments processing

    if Args.similarity != "network":
        Args.X = proc(Args.X, Args)
        Args.Y = proc(Args.Y, Args)
        Args.W = Args.X.T @ Args.Y
        Args.similarity = "network"

    Args.px, Args.py = Args.W.shape

    # Global residualization for kmodularity
    if Args.objective == "kmodularity":
        Args.W = Args.W * (np.sqrt(Args.px * Args.py) / Args.k) / np.sum(np.abs(Args.W))
        Args.W = abct.residualn(Args.W, "degree")

    match Args.objective:
        case "kmodularity": Args.objective = "cokmeans"
        case "kmeans":      Args.objective = "cokmeans"
        case "spectral":    Args.objective = "cospectral"

    if Args.start in ["greedy", "balanced"]:
        Args.DistX = Args.W / np.linalg.norm(Args.W, axis=1, keepdims=True)
        Args.DistY = Args.W / np.linalg.norm(Args.W, axis=0, keepdims=True)
        Args.DistX = 1 - (Args.DistX @ Args.DistX.T)
        Args.DistY = 1 - (Args.DistY.T @ Args.DistY)
    else:
        Args.DistX, Args.DistY = [0], [0]

    return Args


def proc(X, Args):

    # Center data points to mean 0 for covariance and correlation
    if Args.similarity in ["cov", "corr"]:
        X = X - np.mean(X, axis=0)

    # Normalize data points to norm 1 for cosine and correlation
    if Args.similarity in ["cosim", "corr"]:
        X = X / np.linalg.norm(X, axis=0, keepdims=True)
    elif Args.similarity in ["dot", "cov"]:
        X = X / np.sqrt(len(X))

    return X



```
 SHRINKAGE Shrinkage of network or timeseries data

   X1 = shrinkage(X)

   Inputs:
       X:  Network matrix of size n x n, or data matrix of size n x p.
           n is the number of nodes or data points and
           p is the number of features.

   Outputs:
       X1: Shrunken network or timeseries matrix.

   Methodological notes:
       The shrinkage algorithm uses cubic interpolation to "despike" an
       initial eigenspectrum peak.

   See also:
       RESIDUALN.

```



from typing import Literal
from numpy.typing import ArrayLike
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import numpy as np

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def shrinkage(X: ArrayLike) -> np.ndarray:

    X = np.asarray(X)

    if not ((X.shape[0] == X.shape[1]) and np.allclose(X, X.T)):
        raise ValueError("Network matrix must be symmetric.")

    # Compute eigendecomposition
    D, V = np.linalg.eigh(X)
    idx = np.argsort(D)[::-1]
    V = V[:, idx]
    D = D[idx]

    n = len(D)

    # Fit cubic polynomial to all eigenvalues
    bk, r0 = np.polyfit(np.arange(1, n + 1), D, 3, full=True)[:2]
    rms0 = np.sqrt(r0[0] / n)  # get rms

    # Rescale x to [0, 1]
    x = (np.arange(1, n + 1) - 1) / (n - 1)
    y = np.zeros(n)

    # Find optimal fit
    for k in range(n):
        b = bk
        bk, rk = np.polyfit(np.arange(k + 1, n + 1), D[k:], 3, full=True)[:2]
        assert len(np.arange(k + 1, n + 1)) == (n - k)
        rmsk = np.sqrt(rk[0] / (n - k))
        y[k] = (rms0 - rmsk) / rms0
        # Detect knee of optimal fit and break
        if k > 0 and (y[k] - x[k]) < (y[k-1] - x[k-1]):
            break

    # Apply the optimal fit
    D = np.polyval(b, np.arange(1, n + 1))
    return V @ np.diag(D) @ V.T


shrinkage.__doc__ = files("abct").joinpath("docstrings", "doc_shrinkage.py").read_text().replace("```", "")



import warnings
from typing import Literal, Optional
from numpy.typing import ArrayLike
from scipy.sparse import sparray
from pydantic import validate_call, ConfigDict
from importlib.resources import files

import abct
import numpy as np
from scipy import sparse

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def kneicomp(
    W: ArrayLike | sparray,
    k: int,
    weight: Literal["weighted", "binary"] = "weighted",
    **kwargs,
) -> np.ndarray:

    W = np.asarray(W)

    # Default kneighbor arguments
    args = {
        "type": "common",
        "kappa": 0.1,
        "similarity": "network",
        "method": "direct",
    }
    # Update arguments with kwargs
    args.update(kwargs)
    kwargs = {key: kwargs[key] for key in set(kwargs.keys()) - set(args.keys())}

    # Get neighbors matrix
    B = abct.kneighbor(W, args["type"], args["kappa"], args["similarity"], args["method"])

    # Get components
    match weight:
        case "weighted":
            if kwargs:
                warnings.warn("Ignoring Name=Value arguments for weighted components.")
            _, V = sparse.linalg.eigs(B, k=k + 1)
            assert not np.any(np.imag(V)), "Weighted components should be real-valued"
            V = np.real(V)
            return V[:, 1:]  # Remove first eigenvector
        case "binary":
            M = abct.loyvain(B.toarray(), k, "kmodularity", "network", **kwargs)[0]
            V = np.zeros((len(M), k))
            V[np.arange(len(M)), M] = 1
            return V


kneicomp.__doc__ = files("abct").joinpath("docstrings", "doc_kneicomp.py").read_text().replace("```", "")



```
 DISPERSION Dispersion of network matrix

   D = dispersion(W)
   D = dispersion(W, type, M)

   Inputs:
       W: Network matrix of size n x n.

       type: Dispersion type
           "coefvar2": Squared coefficient of variation (default).
           "kpartcoef": k-Participation coefficient.

       M: Module vector of length n (if type is "kpartcoef" only).

   Outputs:
       D: Dispersion vector (length n).

   Methodological notes:
       The squared coefficient of variation, or CV2, is the ratio of the
       variance to the square of the mean. CV2 is equivalent to the ratio
       of the second moment to the square of the first moment.

       The participation coefficient is a popular module-based measure of
       connectional diversity. The k-participation coefficient is the
       participation coefficient normalized by module size.

       CV2 is approximately equivalent to the k-participation coefficient
       in homogeneously modular networks, such as correlation or
       co-neighbor networks.

   See also:
       DEGREE.

```
